import logging

import torch


def check_gpu_memory():
    if torch.cuda.is_available():
        allocated = torch.cuda.memory_allocated(0) / 1024**3  # Convert to GB
        reserved = torch.cuda.memory_reserved(0) / 1024**3  # Convert to GB
        total_memory = (
            torch.cuda.get_device_properties(0).total_memory / 1024**3
        )  # Convert to GB
        free_memory = total_memory - reserved

        logging.info(f"Allocated memory: {allocated:.2f} GB")
        logging.info(f"Reserved memory: {reserved:.2f} GB")
        logging.info(f"Total memory: {total_memory:.2f} GB")
        logging.info(f"Free memory: {free_memory:.2f} GB")

        return free_memory
    else:
        logging.info("CUDA is not available.")
        return 0



def fetch_answers_for_questions(api, question_ids, num_answers=5):
    all_answers = api.fetch(
        "questions/{}/answers".format(";".join(question_ids)),
        sort="votes",
        filter="withbody",
        pagesize=num_answers,
    )
    return all_answers



import logging

from sentence_transformers import SentenceTransformer, util

from tldrai.modules.utils.gpu import check_gpu_memory
from tldrai.modules.utils.logging import configure_logging


class SentenceSimilarity:
    def __init__(
        self,
        model_name="sentence-transformers/all-MiniLM-L6-v2",
        verbose=False,
    ):
        """
        Initializes the SentenceSimilarity class with a specified transformer model.

        Args:
            model_name (str): The name of the sentence transformer model to be used.
            Defaults to 'sentence-transformers/all-MiniLM-L6-v2'.
        """
        self.verbose = verbose
        configure_logging(logging.DEBUG if self.verbose else logging.INFO)
        self.logger = logging.getLogger(__name__)
        self.model_memory_used = 0.5  # GB
        free_memory = check_gpu_memory()

        # None lets it run another internal check
        device = None if free_memory >= self.model_memory_used else "cpu"
        self.model = SentenceTransformer(model_name, device=device)

    def encode_sentences(self, sentences):
        """
        Encodes a list of sentences into embeddings using the loaded model.

        Args:
            sentences (List[str]): A list of sentences to encode.

        Returns:
            np.ndarray: The embeddings of the provided sentences.
        """
        return self.model.encode(sentences)

    def compare_base_to_others(self, base_sentence, other_sentences):
        """
        Compares a base sentence to a list of other sentences to find their similarity scores.

        Args:
            base_sentence (str): The base sentence to compare against.
            other_sentences (List[str]): A list of other sentences to compare to the base sentence.

        Returns:
            np.ndarray: A 2D numpy array of similarity scores between the base sentence and each of the other sentences.
        """
        # Encode the base sentence and other sentences
        base_embedding = self.encode_sentences([base_sentence])
        other_embeddings = self.encode_sentences(other_sentences)

        # Calculate and return the cosine similarity scores
        cosine_scores = util.cos_sim(base_embedding, other_embeddings)
        return cosine_scores



import ast
import logging
import os

import click
from omegaconf import OmegaConf
from tldrai.run import main

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def load_config(config_path="ollama_stable_code.yml"):
    """Load configuration

    Args:
        config_path (str): Path to the configuration file

    Returns:
        OmegaConf: Configuration object
    """
    print("dirname", os.path.dirname(__file__))
    config_path = os.path.join(os.path.dirname(__file__), "..", "config", config_path)

    if not os.path.exists(config_path):
        raise FileNotFoundError(f"Config file not found: {config_path}")
    return OmegaConf.load(config_path)


def apply_overrides(config, overrides):
    """Applies CLI overrides to the configuration

    Args:
        config (OmegaConf): omegaconf configuration to override
        overrides (tuple of str): overrides to apply in the format key=value

    Returns:
        None
    """
    for override in overrides:
        key, value = override.split("=", 1)
        keys = key.split(".")
        cfg_node = config
        for k in keys[:-1]:
            cfg_node = cfg_node[k]
        try:
            parsed_value = ast.literal_eval(value)
        except (ValueError, SyntaxError):
            parsed_value = value
        cfg_node[keys[-1]] = parsed_value


def set_logger_level(verbose):
    """Sets logger level

    Args:
        verbose (bool): Debug if True, Info if False

    Returns:
        None
    """
    if verbose:
        logger.setLevel(logging.DEBUG)
        for handler in logger.handlers:
            handler.setLevel(logging.DEBUG)


@click.command()
@click.argument("question", nargs=-1)
@click.option(
    "--config",
    type=str,
    default="ollama_stable_code.yml",
    help="Name of the config file",
)
@click.option(
    "--set",
    "overrides",
    multiple=True,
    help="Set configuration key-value pairs (key=value)",
)
@click.option("-v", "--verbose", is_flag=True, help="Output all logs verbose")
@click.option(
    "-s",
    "--search",
    is_flag=True,
    default=False,
    help="Search for similar questions on Stack Overflow",
)
def main_cli(question, config, overrides, verbose, search):
    """Invokes the main runner

    Args:
        question (tuple of str): Question to ask the LLM
        config (str): Path to the configuration file
        overrides (tuple of str): Configuration key-value pairs to override
        verbose (bool): Output all logs verbose if True
        search (bool): Search for similar questions on Stack Overflow if True

    Returns:
        None
    """
    config = load_config(config)

    config.question = " ".join(question)
    if overrides:
        apply_overrides(config, overrides)

    config.verbose = verbose
    config.no_search = not search

    set_logger_level(verbose)

    main(config)


if __name__ == "__main__":
    main_cli()



"""
Main running script
"""

import datetime
import logging

import hydra
from omegaconf import DictConfig
from stackapi import StackAPI

from tldrai.modules.pre_inference.pre_summarization import (
    prepare_prompt_for_tokenizer,
    prepare_summarization_input,
)
from tldrai.modules.stack_overflow.fetch import fetch_answers_for_questions
from tldrai.modules.stack_overflow.process import process_answers, process_questions
from tldrai.modules.stack_overflow.search import search_stack_overflow_questions
from tldrai.modules.utils.logging import configure_logging

# Set up logging
logger = logging.getLogger(__name__)


@hydra.main(config_path="../config", config_name="config", version_base=None)
def main(cfg: DictConfig):
    configure_logging(logging.DEBUG if cfg.get("verbose", False) else logging.INFO)

    start_time_ms = round(datetime.datetime.now().timestamp() * 1000)
    logger.debug("Starting process...")

    question = cfg.question
    logger.debug(f"Processing question: {question}")
    logger.debug(f"no_search set to: {cfg.no_search}")

    summarizer = hydra.utils.instantiate(cfg.summarization_pipeline)
    system_prompt = "" if not cfg.prompt else cfg.prompt.format(question)

    if cfg.no_search:
        summarization_input = ""
        fetch_time_ms = process_time_ms = round(
            datetime.datetime.now().timestamp() * 1000
        )
    else:
        SITE = StackAPI("stackoverflow")
        SITE.page_size = cfg.stack_overflow.page_size

        try:
            questions = search_stack_overflow_questions(
                SITE,
                question,
                num_questions=cfg.stack_overflow.num_questions,
                search_style=cfg.stack_overflow.search_style,
                tag=cfg.stack_overflow.tag,
            )
        except RuntimeError:
            logger.error(
                "\tNo questions found for the query. Try rephrasing and searching again. Alternatively, try asking the "
                "LLM without the search."
            )
            return

        questions = process_questions(questions)
        question_ids = [str(x["question_id"]) for x in questions]

        answers = fetch_answers_for_questions(
            SITE, question_ids, num_answers=cfg.stack_overflow.num_answers
        )
        fetch_time_ms = round(datetime.datetime.now().timestamp() * 1000)
        processed_answers = process_answers(answers["items"], questions, question)
        history_len = 0 if cfg.history is None else len(cfg.history)
        max_new_tokens = cfg.get("max_new_tokens", 0)
        summarization_input = prepare_summarization_input(
            processed_answers,
            n=5,
            max_new_tokens=max_new_tokens,
            token_limit=cfg.model_token_limit,
            history_len=history_len,
        )
        process_time_ms = round(datetime.datetime.now().timestamp() * 1000)

    summarization_input = prepare_prompt_for_tokenizer(
        cfg, question, summarization_input, system_prompt
    )

    generation_params = cfg.generation_params
    try:
        summary, input_len, token_shape = summarizer.run(
            summarization_input, **generation_params
        )
        print("\n")
        logger.debug("Summary generated successfully.")
        end_time_ms = round(datetime.datetime.now().timestamp() * 1000)
        status = "success"
        status_message = ""
        token_usage = token_shape[1]
        output = summary[input_len:]
    except Exception as e:
        end_time_ms = round(datetime.datetime.now().timestamp() * 1000)
        status = "error"
        status_message = str(e)
        token_usage = 0
        summary = ""
        output = ""
        input_len = 0

        logger.error(f"Error: {status_message}")
        raise

    run_params = dict(
        **generation_params,
        **cfg.stack_overflow,
        **cfg.summarization_pipeline,
        prompt=cfg.prompt,
        fetch_time_ms=fetch_time_ms - start_time_ms,
        process_time_ms=process_time_ms - fetch_time_ms,
        token_usage=token_usage,
        success=True if status == "success" else False,
        output="" if status == "error" else output,
        question=question,
        status_message=status_message,
        inference_time_ms=end_time_ms - process_time_ms,
        summarization_input=summarization_input,
        prompt_ending=cfg.prompt_ending,
        is_prompt_codified=cfg.is_prompt_codified,
    )

    logger.debug(f"Run parameters: {run_params}")



"""
Searches StackOverflow for questions matching the given query
"""

import logging

from tldrai.modules.utils.exceptions import ArgumentError
from tldrai.modules.utils.logging import configure_logging


def search_stack_overflow_questions(
    api, query, tag="", num_questions=3, search_style="excerpts", verbose=False
):
    """Searches StackOverflow for questions matching the given query

    Args:
        api: StackOverflow API object
        query: query to search
        tag: any special tag to search
        num_questions: number of questions to search
        search_style: "excerpts" or "simple"
        verbose: debug if True, otherwise info

    Returns:
        list(dict()): StackOverflow questions found
    """
    configure_logging(logging.DEBUG if verbose else logging.INFO)
    logger = logging.getLogger(__name__)
    if search_style == "simple":
        if tag:
            questions = api.fetch(
                "search",
                intitle=query,
                tagged=tag,
                sort="relevance",
                pagesize=num_questions,
            )
        else:
            questions = api.fetch(
                "search",
                intitle=query,
                sort="relevance",
                pagesize=num_questions,
            )
    elif search_style == "excerpts":
        questions = api.fetch(
            "search/excerpts",
            q=query,
            sort="relevance",
            pagesize=num_questions,
        )
    else:
        raise ArgumentError(
            f'Unknown search_style {search_style}. Currently "simple" and "excerpts" are supported'
        )
    logger.debug("Questions: ", questions)
    if not questions["items"]:
        raise RuntimeError(
            "No questions found for the query. try rephrasing and searching again"
        )
    return questions["items"]



import itertools
import logging
import threading
import time

import ollama

from tldrai.modules.utils.logging import configure_logging


class OllamaPipeline:
    def __init__(self, model, verbose=False, stream_responses=True, keep_alive="5m"):
        self.model = model
        self.verbose = verbose
        self.stream_responses = stream_responses
        self.keep_alive = keep_alive
        configure_logging(logging.DEBUG if self.verbose else logging.INFO)
        self.logger = logging.getLogger(__name__)
        self.logger.debug(f"Initializing OllamaPipeline with model {self.model}")

        if not self.is_model_pulled(self.model):
            self.pull_model(self.model)

    def is_model_pulled(self, model_name):
        models = ollama.list()["models"]
        if ":" not in model_name:
            model_name += ":latest"
        return any(model["name"] == model_name for model in models)

    def pull_model(self, model_name):
        self.logger.info(f"Pulling model {model_name}...")
        ollama.pull(model_name)
        self.logger.info(f"Model {model_name} pulled successfully.")

    def run(self, prompt, **gen_params):
        if self.stream_responses:
            return self._stream_response(prompt, **gen_params)
        else:
            return self._generate_with_animation(prompt, **gen_params)

    def _stream_response(self, prompt, **gen_params):
        stream = ollama.chat(
            model=self.model,
            messages=prompt,
            stream=True,
            keep_alive=self.keep_alive,
            options=gen_params,
        )
        response = ""
        for chunk in stream:
            print(chunk["message"]["content"], end="", flush=True)
            response += chunk["message"]["content"]
        input_len = sum([len(x["content"]) for x in prompt])
        token_shape = (None, input_len + len(response))
        return response, input_len, token_shape

    def _generate_with_animation(self, prompt, **gen_params):
        done = False

        def animate():
            for c in itertools.cycle(["|", "/", "-", "\\"]):
                if done:
                    break
                print(f"\rGenerating response... {c}", end="", flush=True)
                time.sleep(0.1)

        t = threading.Thread(target=animate)
        t.start()

        response = ollama.chat(
            model=self.model,
            messages=prompt,
            stream=False,
            keep_alive=self.keep_alive,
            options=gen_params,
        )["message"]["content"]

        done = True
        t.join()
        print("\r" + " " * 30, end="\r", flush=True)
        print(response)

        input_len = sum([len(x["content"]) for x in prompt])
        token_shape = (None, input_len + len(response))
        return response, input_len, token_shape



def prepare_summarization_input(
    processed_answers, n=3, max_new_tokens=200, token_limit=2048, history_len=0
):
    combined_text = ""  # "<<<START OF LISTED ANSWERS>>>\n"

    # Combine answers, or select top n
    for count, answer in enumerate(processed_answers):
        if count == n:
            break
        answer_info = [f"{k}: {v}\n" for k, v in answer.items()]
        answer_info = f"Answer {count+1}:\n" + "".join(answer_info)

        if (
            len(combined_text)
            + max_new_tokens * 1.2
            + len("<<<END OF LISTED ANSWERS>>>")
            + len(answer_info)
            + history_len
            > token_limit
        ):
            break

        combined_text += answer_info

    return combined_text


def prepare_prompt_for_tokenizer(cfg, question, summarization_input, system_prompt):
    user_content = summarization_input + cfg.prompt_ending.format(question)

    if cfg.is_prompt_codified:
        if cfg.prompt:
            summarization_input = [
                {"role": "system", "content": system_prompt},
                {"role": "user", "content": user_content},
            ]
        else:
            summarization_input = [{"role": "user", "content": user_content}]

        if cfg.history:
            history = [
                {"role": key, "content": value} for key, value in cfg.history.items()
            ]
            summarization_input = history + summarization_input
    else:
        summarization_input = system_prompt + user_content

        if cfg.history:
            summarization_input = (
                "\n".join([v for k, v in cfg.history.items()])
                + "\n"
                + summarization_input
            )

    return summarization_input



import logging
import re
from datetime import datetime

import numpy as np
import pandas as pd
from bs4 import BeautifulSoup

from tldrai.modules.similarity_scoring.sentence_similarity import SentenceSimilarity
from tldrai.modules.utils.logging import configure_logging


def process_questions(questions):
    ret_questions = []

    if not questions:
        raise RuntimeError("No questions to process")

    for question in questions:
        if question.get("closed_reason") == "Duplicate":
            continue
        ret_questions.append(question)
    return ret_questions


def preprocess_answer(answer):
    return answer
    # Extract code blocks
    code_blocks = re.findall(r"<code>(.*?)</code>", answer, re.DOTALL)
    code_snippets = "\n".join(code_blocks)
    # Prepend code to the answer body to emphasize it
    return f"[Code]\n{code_snippets}\n[/Code]\n\n{answer}"


# Function to extract and merge code blocks
def extract_and_merge_code_blocks(
    html_content, delimiter="```\n```", delimiter_wrap="```"
):
    soup = BeautifulSoup(html_content, "lxml")

    # Find all code blocks
    code_blocks = soup.find_all("code")

    # Extract text from code blocks
    code_texts = ["\n" + code.get_text(strip=True) for code in code_blocks if code]

    # Merge code blocks with a delimiter
    merged_code = delimiter.join(code_texts)

    if delimiter_wrap:
        merged_code = delimiter + merged_code + delimiter

    return merged_code


def process_answers(answers, questions, user_question, verbose=False):
    configure_logging(logging.DEBUG if verbose else logging.INFO)
    logger = logging.getLogger(__name__)
    questions = pd.DataFrame(questions).set_index("question_id")
    answers = pd.DataFrame(answers)
    answers = pd.merge(
        answers, questions["title"], left_on="question_id", right_index=True
    )
    logger.debug(answers)
    similarity = (
        SentenceSimilarity(verbose=verbose).compare_base_to_others(
            base_sentence=user_question,
            other_sentences=answers["title"].tolist(),
        )
    ).numpy()
    answers["similarity"] = similarity[0]
    processed = []
    for _, answer in answers.sort_values(by="similarity", ascending=False).iterrows():
        # Convert creation date from epoch time to a readable format
        if not np.isnan(answer["creation_date"]):
            creation_date = datetime.fromtimestamp(
                int(answer["creation_date"])
            ).strftime("%Y-%m-%d")
        else:
            creation_date = "UNKNOWN"
        if not np.isnan(answer["last_activity_date"]):
            last_activity_date = datetime.fromtimestamp(
                int(answer["last_activity_date"])
            ).strftime("%Y-%m-%d")
        else:
            last_activity_date = "UNKNOWN"
        score = answer["score"]
        question = questions.loc[answer["question_id"], "title"]
        is_accepted = answer.get("is_accepted", False)
        # Example processing, customize based on needs
        processed.append(
            {
                "question_title": question,
                "creation_date": creation_date,
                "last_activity_date": last_activity_date,
                "score": score,
                "is_accepted": is_accepted,
                "code": extract_and_merge_code_blocks(answer["body"]),
                "body": preprocess_answer(answer["body"]),
            }
        )
    return processed



import logging


def configure_logging(level=logging.INFO):
    logging.basicConfig(level=level)
    logger = logging.getLogger()
    logger.setLevel(level)
    for handler in logger.handlers:
        handler.setLevel(level)
    return logger
